#pylint: disable=too-many-return-statements,too-many-branches
import typing as tp

import copy

from colt.error import ConfigurationError
from colt.default_registry import DefaultRegistry
from colt.registrable import Registrable


class ColtBuilder:
    _DEFAULT_TYPEKEY = "@type"
    _DEFAULT_ARGSKEY = "*"

    def __init__(self, typekey: str = None, argskey: str = None) -> None:
        self._typekey = typekey or ColtBuilder._DEFAULT_TYPEKEY
        self._argskey = argskey or ColtBuilder._DEFAULT_ARGSKEY

    def __call__(self, config: tp.Any, cls: tp.Type = None) -> tp.Any:
        return self._build(config, "", cls)

    @staticmethod
    def _remove_optional(annotation: type):
        origin = getattr(annotation, "__origin__", None)
        args = getattr(annotation, "__args__", ())
        if origin == tp.Union and len(args) == 2 and args[1] == type(None):
            return args[0]
        return annotation

    @staticmethod
    def _get_constructor_by_name(name: str,
                                 param_name: str,
                                 annotation: tp.Type = None):
        if annotation and issubclass(annotation, Registrable):
            constructor = annotation.by_name(name)
        else:
            constructor = DefaultRegistry.by_name(name)

        if constructor is None:
            raise ConfigurationError(
                f"type not found error at `{param_name}`: {name}")

        return constructor

    def _construct_with_args(
        self,
        constructor: tp.Callable,
        config: tp.Dict[str, tp.Any],
        param_name: str,
    ) -> tp.Any:
        if not config:
            return constructor()

        args_config = config.pop(self._argskey, [])
        if not isinstance(args_config, (list, tuple)):
            raise ConfigurationError(
                f"args must be a list or tuple: {param_name}")
        args = [
            self._build(val, param_name + f".{self._argskey}.{i}")
            for i, val in enumerate(args_config)
        ]

        if isinstance(constructor, type):
            type_hints = tp.get_type_hints(getattr(constructor, "__init__"))
        else:
            type_hints = tp.get_type_hints(constructor)

        kwargs = {
            key: self._build(val, param_name + f".{key}", type_hints.get(key))
            for key, val in config.items()
        }

        return constructor(*args, **kwargs)

    def _build(
        self,
        config: tp.Any,
        param_name: str,
        annotation: tp.Type = None,
    ) -> tp.Any:
        config = copy.deepcopy(config)

        if annotation is not None:
            annotation = self._remove_optional(annotation)

        if annotation == tp.Any:
            annotation = None

        origin = getattr(annotation, "__origin__", None)
        args = getattr(annotation, "__args__", [])

        if config is None:
            return config

        if origin in (tp.List, list):
            value_cls = args[0] if args else None
            return list(
                self._build(x, param_name + f".{i}", value_cls)
                for i, x in enumerate(config))

        if origin in (tp.Set, set):
            value_cls = args[0] if args else None
            return set(
                self._build(x, param_name + f".{i}", value_cls)
                for i, x in enumerate(config))

        if origin in (tp.Tuple, tuple):
            if not args:
                return tuple(
                    self._build(x, param_name + f".{i}")
                    for i, x in enumerate(config))

            if len(args) == 2 and args[1] == Ellipsis:
                return tuple(
                    self._build(x, param_name + f".{i}", args[0])
                    for i, x in enumerate(config))

            return tuple(
                self._build(value_config, param_name + f".{i}", value_cls)
                for i, (value_config,
                        value_cls) in enumerate(zip(config, args)))

        if origin in (tp.Dict, dict):
            key_cls = args[0] if args else None
            value_cls = args[1] if args else None
            return {
                self._build(key_config, param_name + f".[key:{i}]", key_cls):
                self._build(value_config, param_name + f".{key_config}",
                            value_cls)
                for i, (key_config, value_config) in enumerate(config.items())
            }

        if origin == tp.Union:
            if not args:
                return self._build(config, param_name)

            for value_cls in args:
                try:
                    return self._build(config, param_name, value_cls)
                except (ValueError, TypeError, ConfigurationError,
                        AttributeError):
                    continue

            raise ConfigurationError(
                f"Failed to construct argument {param_name} with type {annotation}"
            )

        if isinstance(config, (list, set, tuple)):
            T = type(config)
            value_cls = args[0] if args else None
            return T(
                self._build(x, param_name + f".{i}", value_cls)
                for i, x in enumerate(config))

        if not isinstance(config, dict):
            if annotation is not None and not isinstance(config, annotation):
                raise UserWarning(
                    f"type mismatch at {param_name}: {annotation}")
            return config

        if annotation is None and self._typekey not in config:
            return {
                key: self._build(val, param_name + f".{key}")
                for key, val in config.items()
            }

        if self._typekey in config:
            class_name = config.pop(self._typekey)
            constructor = self._get_constructor_by_name(
                class_name, param_name, annotation)
        else:
            constructor = origin or annotation  # type: ignore

        return self._construct_with_args(constructor, config, param_name)



from colt.registrable import Registrable


class DefaultRegistry(Registrable):
    """default type store"""


DefaultRegistry.register("bool")(bool)
DefaultRegistry.register("byte")(bytes)
DefaultRegistry.register("bytearray")(bytearray)
DefaultRegistry.register("memoryview")(memoryview)
DefaultRegistry.register("complex")(complex)
DefaultRegistry.register("dict")(dict)
DefaultRegistry.register("float")(float)
DefaultRegistry.register("frozenset")(frozenset)
DefaultRegistry.register("int")(int)
DefaultRegistry.register("list")(list)
DefaultRegistry.register("range")(range)
DefaultRegistry.register("set")(set)
DefaultRegistry.register("str")(str)
DefaultRegistry.register("tuple")(tuple)



import typing as tp

from colt.registrable import Registrable
from colt.default_registry import DefaultRegistry
from colt.builder import ColtBuilder
from colt.utils import import_modules
from colt.version import VERSION as __version__


def register(name: str, constructor: str = None, exist_ok: bool = False):
    def decorator(T: tp.Type):
        DefaultRegistry.register(name, constructor, exist_ok)(T)
        return T

    return decorator


def build(config: tp.Any,
          cls: tp.Type = None,
          typekey: str = None,
          argskey: str = None) -> tp.Any:
    builder = ColtBuilder(typekey, argskey)
    return builder(config, cls)



import typing as tp
import importlib
from collections import defaultdict

from colt.error import ConfigurationError

T = tp.TypeVar("T", bound="Registrable")
Registry = tp.Dict[tp.Type, tp.Dict[str, tp.Tuple[tp.Type, tp.Optional[str]]]]


class Registrable:
    _registry: Registry = defaultdict(dict)

    @classmethod
    def register(cls: tp.Type[T],
                 name: str,
                 constructor: str = None,
                 exist_ok: bool = False):
        registry = Registrable._registry[cls]

        def decorator(subclass: tp.Type[T]):
            if not exist_ok and name in registry:
                raise ValueError(f"type name conflict: {name}")

            if constructor and not hasattr(subclass, constructor):
                raise ValueError(
                    f"constructor {constructor} not found in {subclass}")

            registry[name] = (subclass, constructor)

            return subclass

        return decorator

    @classmethod
    def by_name(cls: tp.Type[T], name: str) -> tp.Callable[..., T]:
        subclass, constructor = cls.resolve_class_name(name)

        if not constructor:
            return subclass

        return tp.cast(tp.Callable[..., T], getattr(subclass, constructor))

    @classmethod
    def resolve_class_name(
            cls: tp.Type[T],
            name: str) -> tp.Tuple[tp.Type[T], tp.Optional[str]]:
        registry = Registrable._registry[cls]

        if name in registry:
            subclass, constructor = registry[name]
            return subclass, constructor

        if "." in name:
            submodule, class_name = name.rsplit(".", 1)

            try:
                module = importlib.import_module(submodule)
            except ModuleNotFoundError as e:
                raise ConfigurationError(
                    f"module {submodule} not found ({name})") from e

            try:
                subclass = getattr(module, class_name)
                constructor = None
                return subclass, constructor
            except AttributeError as e:
                raise ConfigurationError(
                    f"class {class_name} not found in {submodule} ({name})"
                ) from e

        raise ConfigurationError(
            f"{name} is not a registered name for {cls.__name__}. ")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['colt']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'colt',
    'version': '0.6.0',
    'description': 'A configuration utility for Python object.',
    'long_description': 'colt\n===\n\n[![Actions Status](https://github.com/altescy/colt/workflows/build/badge.svg)](https://github.com/altescy/colt)\n[![Python version](https://img.shields.io/pypi/pyversions/colt)](https://github.com/altescy/colt)\n[![pypi version](https://img.shields.io/pypi/v/colt)](https://pypi.org/project/colt/)\n[![license](https://img.shields.io/github/license/altescy/colt)](https://github.com/altescy/colt/blob/master/LICENSE)\n\n## Quick Links\n\n- [Installation](#Installation)\n- [Basic Examples](Examples)\n- [kaggle Titanic Example](https://github.com/altescy/colt/tree/master/examples/titanic)\n\n## Introduction\n\n`colt` is a configuration utility for Python objects.\n`colt` constructs Python objects from a configuration dict which is convertable into JSON.\n(Inspired by [AllenNLP](https://github.com/allenai/allennlp))\n\n\n## Installation\n\n```\npip install colt\n```\n\n## Examples\n\n#### Basic Usage\n\n```python\nimport typing as tp\nimport colt\n\n@colt.register("foo")\nclass Foo:\n    def __init__(self, message: str) -> None:\n        self.message = message\n\n@colt.register("bar")\nclass Bar:\n    def __init__(self, foos: tp.List[Foo]) -> None:\n        self.foos = foos\n\nif __name__ == "__main__":\n    config = {\n        "@type": "bar",  # specify type name with `@type`\n        "foos": [\n            {"message": "hello"},  # type of this is inferred from type-hint\n            {"message": "world"},\n        ]\n    }\n\n    bar = colt.build(config)\n\n    assert isinstance(bar, Bar)\n\n    print(" ".join(foo.message for foo in bar.foos))\n        # => "hello world"\n```\n\n#### `scikit-learn` Configuration\n\n```python\nimport colt\n\nfrom sklearn.datasets import load_iris\nfrom sklearn.model_selection import train_test_split\n\nif __name__ == "__main__":\n    config = {\n        # import types automatically if type name is not registerd\n        "@type": "sklearn.ensemble.VotingClassifier",\n        "estimators": [\n            ("rfc", { "@type": "sklearn.ensemble.RandomForestClassifier",\n                      "n_estimators": 10 }),\n            ("svc", { "@type": "sklearn.svm.SVC",\n                      "gamma": "scale" }),\n        ]\n    }\n\n    X, y = load_iris(return_X_y=True)\n    X_train, X_valid, y_train, y_valid = train_test_split(X, y)\n\n    model = colt.build(config)\n    model.fit(X_train, y_train)\n\n    valid_accuracy = model.score(X_valid, y_valid)\n    print(f"valid_accuracy: {valid_accuracy}")\n```\n\n\n### `Registrable` Class\n\nBy using the `Registrable` class, you can devide namespace into each class.\nIn a following example, `Foo` and `Bar` have different namespaces.\n\n```python\nimport colt\n\nclass Foo(colt.Registrable):\n    pass\n\nclass Bar(colt.Registrable):\n    pass\n\n@Foo.register("baz")\nclass FooBaz(Foo):\n    pass\n\n@Bar.register("baz")\nclass BarBaz(Bar):\n    pass\n\n@colt.register("my_class")\nclass MyClass:\n    def __init__(self, foo: Foo, bar: Bar):\n        self.foo = foo\n        self.bar = bar\n\nif __name__ == "__main__":\n    config = {\n        "@type": "my_class",\n        "foo": {"@type": "baz"},\n        "bar": {"@type": "baz"}\n    }\n\n    obj = colt.build(config)\n\n    assert isinstance(obj.foo, FooBaz)\n    assert isinstance(obj.bar, BarBaz)\n```\n',
    'author': 'altescy',
    'author_email': 'altescy@fastmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/altescy/colt',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import typing as tp
import importlib
import pkgutil
import sys


def import_submodules(package_name: str) -> None:
    """
    original code is here:
    https://github.com/allenai/allennlp/blob/v0.9.0/allennlp/common/util.py
    """
    importlib.invalidate_caches()

    sys.path.append('.')

    # Import at top level
    module = importlib.import_module(package_name)
    path = getattr(module, '__path__', [])
    path_string = '' if not path else path[0]

    for module_finder, name, _ in pkgutil.walk_packages(path):
        if path_string and getattr(module_finder, "path") != path_string:
            continue
        subpackage = f"{package_name}.{name}"
        import_submodules(subpackage)


def import_modules(module_names: tp.List[str]) -> None:
    """
    This method import modules recursively.
    You should call this method to register your classes
    if these classes are written on several files.
    """
    for module_name in module_names:
        import_submodules(module_name)
